from typing import Literal, Type

from bitpapa_pay.methods.base import BaseMethod, BaseOutData
from bitpapa_pay.types.exchange_rates import GetExchangeRatesOut


class GetExchangeRates(BaseMethod):
    @property
    def endpoint(self) -> str:
        return "/api/v1/exchange_rates/all"

    @property
    def request_type(self) -> Literal["GET"]:
        return "GET"

    @property
    def returning_model(self) -> Type[GetExchangeRatesOut]:
        return GetExchangeRatesOut

    def get_data(self) -> BaseOutData:
        return BaseOutData(
            endpoint=self.endpoint,
            request_type=self.request_type,
            returning_model=self.returning_model
        )



from typing import List, Optional, Union

from pydantic import BaseModel, computed_field


class TelegramInvoice(BaseModel):
    id: str
    currency_code: str
    amount: Union[int, float]
    status: str
    created_at: str
    updated_at: str

    @computed_field
    def url(self) -> str:
        return f"https://t.me/bitpapa_bot?start={self.id}"


class TelegramInvoices(BaseModel):
    invoices: List[TelegramInvoice]


class TelegramInvoiceInputData(BaseModel):
    currency_code: str
    amount: Union[int, float]
    crypto_address: Optional[str]


class CreateTelegramInvoiceInputData(BaseModel):
    api_token: str
    invoice: TelegramInvoiceInputData


class CreateTelegramInvoiceOutputData(BaseModel):
    invoice: TelegramInvoice


class GetTelegramInvoicesInputParams(BaseModel):
    api_token: str



from typing import Any, Literal, Optional

from pydantic import BaseModel


class BaseOutData(BaseModel):
    endpoint: str
    request_type: Literal["GET", "POST"]
    params: Optional[dict] = None
    json_data: Optional[dict] = None
    returning_model: Any



from typing import Literal, Type, Union, Optional

from bitpapa_pay.methods.base import BaseMethod, BaseOutData
from bitpapa_pay.types.telegram import (CreateTelegramInvoiceInputData,
                                        CreateTelegramInvoiceOutputData,
                                        GetTelegramInvoicesInputParams,
                                        TelegramInvoiceInputData,
                                        TelegramInvoices)


class CreateTelegramInvoice(BaseMethod):
    def __init__(
        self,
        api_token: str,
        currency_code: str,
        amount: Union[int, float],
        crypto_address: Optional[str] = None
    ) -> None:
        self.api_token = api_token
        self.currency_code = currency_code
        self.amount = amount
        self.crypto_address = crypto_address

    @property
    def endpoint(self) -> str:
        return "/api/v1/invoices/public"

    @property
    def request_type(self) -> Literal["POST"]:
        return "POST"

    @property
    def returning_model(self) -> Type[CreateTelegramInvoiceOutputData]:
        return CreateTelegramInvoiceOutputData

    def get_data(self) -> BaseOutData:
        return BaseOutData(
            endpoint=self.endpoint,
            request_type=self.request_type,
            json_data=CreateTelegramInvoiceInputData(
                api_token=self.api_token,
                invoice=TelegramInvoiceInputData(
                    currency_code=self.currency_code,
                    amount=self.amount,
                    crypto_address=self.crypto_address
                ),
            ).model_dump(),
            returning_model=self.returning_model
        )


class GetTelegramInvoices(BaseMethod):
    def __init__(self, api_token: str) -> None:
        self.api_token = api_token

    @property
    def endpoint(self) -> str:
        return "/api/v1/invoices/public"

    @property
    def request_type(self) -> Literal["GET"]:
        return "GET"

    @property
    def returning_model(self) -> Type[TelegramInvoices]:
        return TelegramInvoices

    def get_data(self) -> BaseOutData:
        return BaseOutData(
            endpoint=self.endpoint,
            request_type=self.request_type,
            params=GetTelegramInvoicesInputParams(
                api_token=self.api_token
            ).model_dump(),
            returning_model=self.returning_model
        )



from abc import ABC, abstractmethod
from typing import Any, Literal

from bitpapa_pay.types.base import BaseOutData


class BaseMethod(ABC):
    @property
    @abstractmethod
    def endpoint(self) -> str:
        pass

    @property
    @abstractmethod
    def request_type(self) -> Literal["GET", "POST"]:
        pass

    @property
    @abstractmethod
    def returning_model(self) -> Any:
        pass

    def set_params(self, params: dict) -> dict:
        return {key: params[key] for key in params if params[key] is not None}

    @abstractmethod
    def get_data(self) -> BaseOutData:
        pass



from typing import Optional, Union

from aiohttp import ClientSession
from loguru import logger

from bitpapa_pay.methods.addresses import (CreateAddress,
                                           CreateAddressOutputData,
                                           GetAddresses,
                                           GetAddressesOutputData,
                                           GetAddressTransactions,
                                           GetTransactions,
                                           GetTransactionsOutputData)
from bitpapa_pay.methods.base import BaseMethod
from bitpapa_pay.methods.exchange_rates import (GetExchangeRates,
                                                GetExchangeRatesOut)
from bitpapa_pay.methods.telegram import (CreateTelegramInvoice,
                                          CreateTelegramInvoiceOutputData,
                                          GetTelegramInvoices,
                                          TelegramInvoices)
from bitpapa_pay.version import VERSION


class HttpClient:
    def __init__(self, api_token: str, debug: bool = False) -> None:
        self._debug = debug
        self._base_url = "https://bitpapa.com"
        self._api_token = api_token
        self._session: Optional[ClientSession] = None

    def debug_message(self, message: str):
        if self._debug:
            logger.debug(message)

    def get_headers(self):
        return {
            "Content-Type": "application/json",
            "User-Agent": f"AioBitpapaPay/{VERSION}",
            "X-Access-Token": self._api_token
        }

    def get_session(self) -> ClientSession:
        headers = self.get_headers()
        self.debug_message(f"request headers: {headers}")

        if isinstance(self._session, ClientSession):
            return self._session
        self._session = ClientSession(base_url=self._base_url, headers=headers)
        return self._session

    async def close(self):
        if isinstance(self._session, ClientSession):
            await self._session.close()

    async def _get_request(
        self,
        session: ClientSession,
        endpoint: str,
        params: Optional[dict] = None
    ):
        async with session.get(url=endpoint, params=params) as resp:
            return await resp.json()

    async def _post_request(
        self,
        session: ClientSession,
        endpoint: str,
        json_data: Optional[dict] = None
    ):
        async with session.post(url=endpoint, json=json_data) as resp:
            return await resp.json()

    async def _make_request(self, method: BaseMethod):
        session = self.get_session()
        request_data = method.get_data()
        self.debug_message(f"request data: {request_data}")
        self.debug_message(
            f"request url: {self._base_url}{request_data.endpoint}"
        )
        if request_data.request_type == "GET":
            result = await self._get_request(
                session=session,
                endpoint=request_data.endpoint,
                params=request_data.params
            )
        elif request_data.request_type == "POST":
            result = await self._post_request(
                session=session,
                endpoint=request_data.endpoint,
                json_data=request_data.json_data
            )
        self.debug_message(f"request result: {result}")
        return result


class DefaultApiClient(HttpClient):
    async def get_exchange_rates_all(self) -> GetExchangeRatesOut:
        """Get all exchange rates, https://apidocs.bitpapa.com/docs/backend-apis-english/97573257c4827-get-a-v-1-exchange-rate-all

        Returns:
            GetExchangeRatesOut: An object where the keys are abbreviations of 
            a pair of exchange rates separated by "_"
        """
        method = GetExchangeRates()
        result = await self._make_request(method)
        return method.returning_model(**result)


class AdressesApiClient(HttpClient):
    async def get_addresses(
        self,
        currency: Optional[str] = None,
        label: Optional[str] = None
    ) -> GetAddressesOutputData:
        method = GetAddresses(currency, label)
        result = await self._make_request(method)
        return method.returning_model(addresses=result)

    async def create_address(
        self,
        currency: str,
        network: str,
        label: str = ""
    ) -> CreateAddressOutputData:
        method = CreateAddress(currency, network, label)
        result = await self._make_request(method)
        return method.returning_model(**result)

    async def get_transactions(
        self
    ) -> GetTransactionsOutputData:
        method = GetTransactions()
        result = await self._make_request(method)
        return method.returning_model(transactions=result)

    async def get_address_transactions(
        self,
        uuid: str
    ) -> GetTransactionsOutputData:
        method = GetAddressTransactions(uuid)
        result = await self._make_request(method)
        return method.returning_model(**result)


class BitpapaPayClient(HttpClient):
    async def get_invoices(self) -> TelegramInvoices:
        """Get the list of invoices, https://apidocs.bitpapa.com/docs/backend-apis-english/qph49kfhdjx0x-get-the-list-of-invoices

        Returns:
            TelegramInvoices: list of telegram invoices
        """
        method = GetTelegramInvoices(api_token=self._api_token)
        result = await self._make_request(method)
        return method.returning_model(**result)

    async def create_invoice(
        self,
        currency_code: str,
        amount: Union[int, float],
        crypto_address: Optional[str] = None
    ) -> CreateTelegramInvoiceOutputData:
        """Issue an invoice to get payment, https://apidocs.bitpapa.com/docs/backend-apis-english/23oj83o5x2su2-issue-an-invoice
        Args:
            currency_code (str): The ticker of accepted cryptocurrency, example: USDT
            amount (Union[int, float]): The amount in cryptocurrency, example 100

        Returns:
            CreateTelegramInvoiceOutputData: Created telegram invoice data
        """
        method = CreateTelegramInvoice(
            api_token=self._api_token,
            currency_code=currency_code,
            amount=amount,
            crypto_address=crypto_address
        )
        result = await self._make_request(method)
        return method.returning_model(**result)


class BitpapaPay(BitpapaPayClient, DefaultApiClient, AdressesApiClient):
    pass



from typing import List, Literal, Optional, Type, Union
from uuid import UUID

from pydantic import BaseModel

from bitpapa_pay.methods.base import BaseMethod, BaseOutData


class Address(BaseModel):
    id: UUID
    address: Optional[str]
    currency: str
    network: str
    balance: Optional[Union[int, float]]
    label: str


class GetAddressesOutputData(BaseModel):
    addresses: List[Address]


class GetAddressesParams(BaseModel):
    currency: Optional[str] = None
    label: Optional[str] = None


class CreateAddressInputData(BaseModel):
    currency: str
    network: str
    label: str


class CreateAddressOutputData(BaseModel):
    address: Address


class GetTransactionsOutputData(BaseModel):
    transactions: List


class GetAddresses(BaseMethod):
    def __init__(
        self,
        currency: Optional[str] = None,
        label: Optional[str] = None
    ) -> None:
        self.currency = currency
        self.label = label

    @property
    def endpoint(self) -> str:
        return "/a3s/v1/addresses"

    @property
    def request_type(self) -> Literal["GET"]:
        return "GET"

    @property
    def returning_model(self) -> Type[GetAddressesOutputData]:
        return GetAddressesOutputData

    def get_data(self) -> BaseOutData:
        return BaseOutData(
            endpoint=self.endpoint,
            request_type=self.request_type,
            params=self.set_params(GetAddressesParams(
                currency=self.currency,
                label=self.label
            ).model_dump()),
            returning_model=self.returning_model
        )


class CreateAddress(BaseMethod):
    def __init__(
        self,
        currency: str,
        network: str,
        label: str = "",
    ) -> None:
        self.currency = currency
        self.label = label
        self.network = network

    @property
    def endpoint(self) -> str:
        return "/a3s/v1/addresses/new"

    @property
    def request_type(self) -> Literal["POST"]:
        return "POST"

    @property
    def returning_model(self) -> Type[CreateAddressOutputData]:
        return CreateAddressOutputData

    def get_data(self) -> BaseOutData:
        return BaseOutData(
            endpoint=self.endpoint,
            request_type=self.request_type,
            json_data=CreateAddressInputData(
                currency=self.currency,
                network=self.network,
                label=self.label
            ).model_dump(),
            returning_model=self.returning_model
        )


class GetTransactions(BaseMethod):
    @property
    def endpoint(self) -> str:
        return "/a3s/v1/transactions"

    @property
    def request_type(self) -> Literal["GET"]:
        return "GET"

    @property
    def returning_model(self) -> Type[GetTransactionsOutputData]:
        return GetTransactionsOutputData

    def get_data(self) -> BaseOutData:
        return BaseOutData(
            endpoint=self.endpoint,
            request_type=self.request_type,
            returning_model=self.returning_model
        )


class GetAddressTransactions(BaseMethod):
    def __init__(self, uuid: str) -> None:
        self.uuid = uuid

    @property
    def endpoint(self) -> str:
        return f"/a3s/v1/address/{self.uuid}/transactions"

    @property
    def request_type(self) -> Literal["GET"]:
        return "GET"

    @property
    def returning_model(self) -> Type[GetTransactionsOutputData]:
        return GetTransactionsOutputData

    def get_data(self) -> BaseOutData:
        return BaseOutData(
            endpoint=self.endpoint,
            request_type=self.request_type,
            returning_model=self.returning_model
        )
